from __future__ import print_function
import os
import re
import sys
import ujson
import multiprocessing
from multiprocessing import Process

import flask
import random
import shutil
import traceback
import flask_classful
from flask import escape
import flaskext.markdown as flask_markdown
from gevent.server import _tcp_listener
from gevent.wsgi import WSGIServer

from CYLGame.Comp import create_room
from .Game import GameRunner, GridGame, int2base
from .Player import Room
from .Game import GameLanguage
from .Game import average
from .Database import GameDB


ANONYMOUS_SCHOOL = "S00000000"
ANONYMOUS_USER = "00000000"


def static_file(filename):
    resource_path = os.path.join(os.path.split(__file__)[0], "static", filename)
    return resource_path


def get_public_ip():
    # This is taken from: http://stackoverflow.com/a/1267524
    import socket
    return [l for l in ([ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith("127.")][:1], [[(s.connect(('8.8.8.8', 53)), s.getsockname()[0], s.close()) for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]]) if l][0][0]


def find_name_from_code(code):
    import re
    name = re.findall(r"#\s*name:\s*(.*)\s*", code, re.IGNORECASE)
    if len(name):
        return name[0]
    else:
        return ""


class GameServer(flask_classful.FlaskView):
    game = None
    host = None
    port = None
    compression = None
    language = None
    avg_game_count = None
    _avg_game_func = None
    charset = None
    gamedb = None
    route_base = '/'

    @classmethod
    def __load_language(cls):
        if cls.language == GameLanguage.LITTLEPY:
            from littlepython import Compiler
            cls.compiler = Compiler()
        else:
            raise Exception("Invalid language. Could not load hookers.")

    @classmethod
    def __verify_game(cls):
        # TODO: add asserts for every attr used later.
        assert hasattr(cls.game, "GAME_TITLE")
        score_op = getattr(cls.game, "get_score", None)
        assert callable(score_op)

    @classmethod
    def __copy_in_charset(cls, charset):
        if not os.path.exists(static_file("fonts")):
            os.mkdir(static_file("fonts"))
        file_ending = os.path.split(charset)[-1]
        prepostfix = "_col"
        if "_ro." in file_ending:
            prepostfix = "_ro"
        elif "_tc." in file_ending:
            prepostfix = "_tc"
        postfix = file_ending.split(".")[-1]
        def new_token():
            return "".join([random.choice("0123456789ABCDEF") for _ in range(10)]) + prepostfix + "." + postfix
        token = new_token()
        while os.path.exists(static_file(os.path.join("fonts", token))):
            token = new_token()
        new_charset_path = static_file(os.path.join("fonts", token))
        shutil.copyfile(charset, new_charset_path)
        return os.path.split(new_charset_path)[-1]

    @flask_classful.route('/scoreboard', methods=["POST"])
    def scoreboard(self):
        token = flask.request.get_json(silent=True).get('token', '')
        if not self.gamedb.is_user_token(token):
            return flask.jsonify(error="Invalid Token")
        school_tk = self.gamedb.get_school_for_token(token)
        obj = {"school": self.gamedb.get_name(school_tk), "scores": []}
        for user_tk in self.gamedb.get_tokens_for_school(school_tk):
            score = self.gamedb.get_avg_score(user_tk)
            if score is not None:
                obj["scores"] += [{"name": escape(self.gamedb.get_name(user_tk)), "score": self.gamedb.get_avg_score(user_tk)}]
        return ujson.dumps(obj)

    @flask_classful.route('/comp_scoreboards', methods=["POST"])
    def comp_scoreboards(self):
        token = flask.request.get_json(silent=True).get('token', '')
        if not self.gamedb.is_user_token(token):
            return flask.jsonify(error="Invalid Token")
        comps = self.gamedb.get_comps_for_token(token)
        obj = {"comps": []}
        for comp in comps:
            comp_obj = {"scores": []}
            for school in self.gamedb.get_schools_in_comp(comp):
                score = self.gamedb.get_comp_avg_score(comp, school)
                if score is not None:
                    comp_obj["scores"] += [{"name": escape(self.gamedb.get_name(school)), "score": score}]
            obj["comps"] += [comp_obj]
        return ujson.dumps(obj)

    @flask_classful.route('/save_code', methods=['POST'])
    def save_code(self):
        code = flask.request.get_json(silent=True).get('code', '')
        token = flask.request.get_json(silent=True).get('token', '')
        options = flask.request.get_json(silent=True).get('options', None)
        if not self.gamedb.is_user_token(token):
            return flask.jsonify(error="Invalid Token")

        self.gamedb.save_code(token, code, options)
        name = find_name_from_code(code)
        if name:
            self.gamedb.save_name(token, name)

        return flask.jsonify(message="success")

    @flask_classful.route('/game/<gtoken>')
    def get_game_data(self, gtoken):
        if not self.gamedb.is_game_token(gtoken):
            return flask.jsonify(error="Invalid Game Token")

        return ujson.dumps(self.gamedb.get_game_frames(gtoken))

    @flask_classful.route('/game/<gtoken>/<token>')
    def get_player_game_data(self, gtoken, token):
        if not self.gamedb.is_game_token(gtoken):
            return flask.jsonify(error="Invalid Game Token")
        if not self.gamedb.is_user_token(token):
            return flask.jsonify(error="Invalid User Token")

        data = self.gamedb.get_game_frames(gtoken)
        data["player"] = self.gamedb.get_player_game_data(gtoken, token)

        if token == ANONYMOUS_USER:
            self.gamedb.delete_game(gtoken)

        return ujson.dumps(data)

    @flask_classful.route('/sim_avg', methods=['POST'])
    def sim_avg(self):
        # TODO: create this to run the game 100 times returning the average score to the user.
        code = flask.request.get_json(silent=True).get('code', '')
        token = flask.request.get_json(silent=True).get('token', '')
        options = flask.request.get_json(silent=True).get('options', None)
        if not self.gamedb.is_user_token(token):
            return flask.jsonify(error="Invalid Token")
        try:
            prog = self.compiler.compile(code)
        except:
            return flask.jsonify(error="Code did not compile")
        if self.game.MULTIPLAYER:
            self.gamedb.save_code(token, code, options)
            name = find_name_from_code(code)
            if name:
                self.gamedb.save_name(token, name)
            return flask.jsonify(score="still being computed. Check scoreboard later.")
        room = Room([prog])
        runner = GameRunner(self.game, room)
        try:
            score = runner.run_for_avg_score(times=self.avg_game_count, func=self._avg_game_func)
            self.gamedb.save_avg_score(token, score)
            self.gamedb.save_code(token, code, options)
            name = find_name_from_code(code)
            if name:
                self.gamedb.save_name(token, name)
            return flask.jsonify(score=score)
        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            return flask.jsonify(error="Your bot ran into an error at runtime.\n"
                                       "If you think that your bot is correct, please file a bug report!\n"
                                       "Make sure to include your code.")

    @flask_classful.route('/sim', methods=['POST'])
    def sim(self):
        code = flask.request.get_json(silent=True).get('code', '')
        seed_str = flask.request.get_json(silent=True).get('seed', '')
        options = flask.request.get_json(silent=True).get('options', None)
        if options is None:
            options = {}
        seed = random.randint(0, sys.maxsize)
        if seed_str:
            try:
                seed = int(seed_str, 36)
            except:
                return flask.jsonify(error="Invalid Seed")
        try:
            prog = self.compiler.compile(code)
            prog.options = options
            prog.token = "00000000"  # anonymous user
        except:
            return flask.jsonify(error="Code did not compile")
        room = Room([prog])
        if self.game.MULTIPLAYER:
            computer_bot_class = self.game.default_prog_for_computer()
            players = []
            for _ in range(self.game.get_number_of_players() - 1):
                players += [computer_bot_class()]
            room = Room([prog] + players)
        runner = GameRunner(self.game, room)
        try:
            runner.run_for_playback(seed=seed)
            game_data = {"screen": room.screen_cap,
                         "seed": int2base(seed, 36)}
            player_data = {}
            for player in room.bots:
                if hasattr(player, "token"):
                    player_data[player.token] = room.debug_vars[player]
            gtoken = self.gamedb.add_new_game(game_data, per_player_data=player_data)
            result = ujson.dumps({"gtoken": gtoken})
            # result = ujson.dumps(res)
        except Exception:
            traceback.print_exc(file=sys.stdout)
            return flask.jsonify(error="Your bot ran into an error at runtime.\n"
                                       "If you think that your bot is correct, please file a bug report!\n"
                                       "Make sure to include your code.")
        return result

    @flask_classful.route('/check_token', methods=['POST'])
    def check_token(self):
        token = flask.request.get_json(silent=True).get('token', '')
        if not self.gamedb.is_user_token(token):
            return flask.jsonify(error="Invalid Token")
        else:
            return flask.jsonify(result=True)

    @flask_classful.route('/load_code', methods=['POST'])
    def load_code(self):
        token = flask.request.get_json(silent=True).get('token', '')
        if not self.gamedb.is_user_token(token):
            return flask.jsonify(error="Invalid Token")
        else:
            code, options = self.gamedb.get_code_and_options(token)
            return flask.jsonify(code=code, options=options)

    def index(self):
        intro = self.game.get_intro() + GameLanguage.get_language_description(self.language)
        if self.game.GRID:
            return flask.render_template('grid.html', game_title=self.game.GAME_TITLE,
                                    example_bot=self.game.default_prog_for_bot(self.language), char_width=self.game.CHAR_WIDTH,
                                    char_height=self.game.CHAR_HEIGHT, screen_width=self.game.SCREEN_WIDTH,
                                    screen_height=self.game.SCREEN_HEIGHT, char_set=self.charset,
                                    intro_text=intro)
        else:
            return flask.render_template('nongrid.html', game_title=self.game.GAME_TITLE,
                                    example_bot=self.game.default_prog_for_bot(self.language), 
                                    screen_width=self.game.SCREEN_WIDTH,
                                    screen_height=self.game.SCREEN_HEIGHT,
                                    intro_text=intro, options=self.game.OPTIONS)

    @classmethod
    def serve(cls, game, host='', port=5000, compression=False, language=GameLanguage.LITTLEPY,
              avg_game_count=10, num_of_threads=None, game_data_path="temp_game", avg_game_func=average,
              debug=True):
        cls.game = game
        cls.host = host
        cls.port = port
        cls.compression = compression
        cls.language = language
        cls.avg_game_count = avg_game_count
        cls._avg_game_func = avg_game_func
        cls.gamedb = GameDB(game_data_path)
        # setup anonymous school with an anonymous user
        if not cls.gamedb.is_school_token(ANONYMOUS_SCHOOL):
            cls.gamedb.add_new_school(_token=ANONYMOUS_SCHOOL)
        if not cls.gamedb.is_user_token(ANONYMOUS_USER):
            cls.gamedb.get_new_token(ANONYMOUS_SCHOOL, _token=ANONYMOUS_USER)

        if issubclass(game, GridGame):
            cls.charset = cls.__copy_in_charset(game.CHAR_SET)

        cls.app = flask.Flask(__name__.split('.')[0])

        @cls.app.template_filter('markdown')
        def markdown_filter(data):
            from flask import Markup
            from markdown import markdown

            return Markup(markdown(data))

        if cls.compression:
            # BUUT WHAT ABOUT BREACH ATTACKS
            import flask_compress
            flask_compress.Compress(cls.app)
        cls.register(cls.app)
        cls.__load_language()

        print("Starting server at {}:{}".format(cls.host, cls.port))

        if debug:
            cls.app.run(cls.host, cls.port)
        else:
            listener = _tcp_listener((cls.host, cls.port))

            def serve_forever(listener):
                WSGIServer(listener, cls.app).serve_forever()

            if num_of_threads is None:
                num_of_threads = multiprocessing.cpu_count()

            for i in range(num_of_threads):
                Process(target=serve_forever, args=(listener,)).start()

            serve_forever(listener)

        print("Dying...")
        if cls.charset and os.path.exists(static_file(os.path.join("fonts", cls.charset))):
            print("Removing charset...")
            os.remove(static_file(os.path.join("fonts", cls.charset)))
        print("All good :)")


# This is for backwards compatibility
serve = GameServer.serve



from __future__ import print_function
from .Game import GameRunner
from .Player import Room
from random import random, choice


def create_room(gamedb, bot, compiler, size):
    stokens = gamedb.get_school_tokens()
    pool = []
    for stoken in stokens:
        pool += gamedb.get_tokens_for_school(stoken)

    if len(pool) == 0:
        from copy import deepcopy
        return Room([bot] + [deepcopy(bot) for _ in range(size - 1)])

    bots = [bot]
    while True:
        token = choice(pool)
        code, options = gamedb.get_code_and_options(token)
        try:
            prog = compiler.compile(code)
            prog.options = options
            bots += [prog]
            if len(bots) >= size:
                return Room(bots)
        except:
            print("Couldn't compile code for '{}' in '{}'".format(token, gamedb.get_school_for_token(token)))


def create_room_for_school(gamedb, stoken):
    # TODO: write this.
    raise Exception("Not Implemented!")


def avg(scores):
    return float((sum(scores) * 100) / len(scores)) / 100


def sim_competition(compiler, game, gamedb, token, runs, debug=False, score_func=avg):
    assert gamedb is not None
    assert gamedb.is_comp_token(token)

    seeds = [random() for _ in xrange(2*runs+5)]

    for school in gamedb.get_schools_in_comp(token):
        if debug:
            print("Got school '" + school + "'")
        max_score = 0
        max_code = ""
        for student in gamedb.get_tokens_for_school(school):
            if debug:
                print("Got student '" + student + "'")
            code = gamedb.get_code(student)
            if not code:
                continue
            if debug:
                print("Compiling code...")
            prog = compiler.compile(code.split("\n"))
            if debug:
                print("Setting up game runner...")
            runner = GameRunner(game, prog)
            if debug:
                print("Simulating...")
            score = None

            scores = []
            count = 0
            seed = 0
            # TODO: make this able to run in a pool of threads (So it can be run on multiple CPUs)
            while count < runs:
                try:
                    if seed >= len(seeds):
                        print("Ran out of seeds")
                        break
                    scores += [runner.run_for_avg_score(times=1, seed=seeds[seed])]
                    # print(scores[-1])
                    # import sys
                    # sys.stdout.flush()
                    count += 1
                    seed += 1
                except Exception as e:
                    print("There was an error simulating the game (Moving to next seed):", e)
                    seed += 1
            score = score_func(scores)
            # score = runner.run_for_avg_score(times=runs)

            # while score is None:
            #     try:
            #     except Exception as e:
            #         print("There was an error simulating the game:", e)
            if score > max_score:
                max_score = score
                max_code = code
        if debug:
            print("Saving score...", max_score)
        gamedb.set_comp_avg_score(token, school, max_score)
        gamedb.set_comp_school_code(token, school, max_code)
    if debug:
        print("All done :)")



from __future__ import division
import os.path
import random
import string
import sys

from CYLGame.Frame import GridFrameBuffer
from CYLGame.Player import UserProg

FPS = 30


# From: http://stackoverflow.com/a/2267446/4441526
digs = string.digits + string.ascii_letters
def int2base(x, base):
    if x < 0:
        sign = -1
    elif x == 0:
        return digs[0]
    else:
        sign = 1

    x *= sign
    digits = []

    while x:
        digits.append(digs[x % base])
        x //= base

    if sign < 0:
        digits.append('-')

    digits.reverse()

    return ''.join(digits)


def scorer(func):
    """This function is a decorator for a scoring function.
       This is hack a to get around self being passed as the first argument to the scoring function."""
    def wrapped(a, b=None):
        if b is not None:
            return func(b)
        return func(a)
    return wrapped


@scorer
def average(scores):
    return float((sum(scores) * 100) / len(scores)) / 100


def data_file(filename):
    resource_path = os.path.join(os.path.split(__file__)[0], "data", filename)
    return resource_path


class GameLanguage(object):
    LITTLEPY = 0

    @staticmethod
    def get_language_description(language):
        if language == GameLanguage.LITTLEPY:
            return open(data_file("little_python_intro.md")).read()


class Game(object):
    WEBONLY = True
    SCREEN_WIDTH = 0
    SCREEN_HEIGHT = 0
    GAME_TITLE = ""
    OPTIONS = None
    MULTIPLAYER = False  # TODO: document
    TURN_BASED = False  # TODO: document

    def is_running(self):
        """This is how the game runner knows if the game is over.

        Returns:
            bool: True if the game should still be play. False otherwise
        """
        raise Exception("Not implemented!")

    def create_new_player(self, prog):
        """This creates n new objects that inherits from the Player class.

        Returns:
            n new objects that inherit from the Player class with the given program.
        """
        raise Exception("Not implemented!")

    def get_debug_vars(self):
        raise Exception("Not implemented!")

    def do_turn(self):
        raise Exception("Not implemented!")

    def get_frame(self):
        raise Exception("Not implemented!")

    def init_board(self):
        raise Exception("Not implemented!")

    @staticmethod
    def default_prog_for_bot(language):
        raise Exception("Not implemented!")

    @staticmethod
    def default_prog_for_computer():
        """This method is only for multi-player games to implement.
        Returns:
            type[Prog]
        """
        raise Exception("Not implemented!")

    @staticmethod
    def get_intro():
        raise Exception("Not implemented!")

    @staticmethod
    def get_move_consts():
        return ConstMapping()

    @staticmethod
    def get_number_of_players():
        """This method is only for multi-player games to implement.
        Returns:
            int: The number of players needed to play the game.
        """
        raise Exception("Not implemented!")


class NonGridGame(Game):
    WEBONLY = True
    GRID = False

    def read_bot_state(self, state):
        raise Exception("Not Implemented!")

    def get_vars_for_bot(self):
        raise Exception("Not Implemented!")


class ConstMapping(dict):
    def __init__(self, seq=None, **kwargs):
        """
        dict() -> new empty dictionary
        dict(mapping) -> new dictionary initialized from a mapping object's
            (key, value) pairs
        dict(iterable) -> new dictionary initialized as if via:
            d = {}
            for k, v in iterable:
                d[k] = v
        dict(**kwargs) -> new dictionary initialized with the name=value pairs
            in the keyword argument list.  For example:  dict(one=1, two=2)
        # (copied from class doc)
        """
        super(ConstMapping, self).__init__()
        if isinstance(seq, dict):
            for v, k in seq.items():
                self[v] = k
        elif kwargs:
            for v, k in kwargs:
                self[v] = k
        else:
            for v, k in seq:
                self[v] = k

    def __setitem__(self, key, value):
        if key in self:
            del self[key]
        if value in self:
            del self[value]
        dict.__setitem__(self, key, value)
        dict.__setitem__(self, value, key)

    def __delitem__(self, key):
        dict.__delitem__(self, self[key])
        dict.__delitem__(self, key)

    def __len__(self):
        return dict.__len__(self) // 2


class GridGame(Game):
    WEBONLY = False
    GRID = True
    SCREEN_WIDTH = 80
    SCREEN_HEIGHT = 25
    CHAR_WIDTH = 8
    CHAR_HEIGHT = 8
    GAME_TITLE = ""
    CHAR_SET = data_file("fonts/terminal8x8_gs_ro.png")
    __frame_buffer = None

    def is_running(self):
        """This is how the game runner knows if the game is over.

        Returns:
            bool: True if the game should still be play. False otherwise
        """
        raise Exception("Not implemented!")

    def create_new_player(self, prog):  # TODO: add option to create computer, user or bot players.
        """ TODO: write this
        """
        raise Exception("Not Implemented!")

    def do_turn(self):
        """This function should read the new state of all the players and react to them."""
        raise Exception("Not implemented!")

    def draw_screen(self, frame_buffer):
        """WARNING: There MUST NOT be any game logic in this function since it isn't called when simulating the game
                    during the competitions.
        """
        raise Exception("Not implemented!")

    def get_frame(self):
        if self.__frame_buffer is None:
            self.__frame_buffer = GridFrameBuffer(self.SCREEN_WIDTH, self.SCREEN_HEIGHT)
        self.draw_screen(self.__frame_buffer)
        return self.__frame_buffer.dump()

    def get_vars(self, player):
        """ TODO: write this
        """
        raise Exception("Not implemented!")

    def get_score(self):
        """This is the game runner gets the ending or mid-game score

        Returns:
            int: The current score.
        """
        raise Exception("Not implemented!")

    @staticmethod
    def default_prog_for_bot(language):
        raise Exception("Not implemented!")

    @staticmethod
    def get_intro():
        raise Exception("Not implemented!")

    @staticmethod
    def get_move_consts():
        return ConstMapping({"north": ord("w"), "south": ord("s"), "west": ord("a"), "east": ord("d"),
                "northeast": ord("e"), "southeast": ord("c"), "northwest": ord("q"), "southwest": ord("z")})


class GameRunner(object):
    def __init__(self, game_class, room=None):
        self.game_class = game_class  # type: Type[Game]
        self.user_is_playing = False
        if room is None:
            self.user_is_playing = True
        else:
            self.room = room  # type: Room
        self.players = []

    def __run_for(self, score=False, playback=False, seed=None):
        assert len(self.room.bots) > 0  # Make sure that we have a bot to run
        assert score != playback

        if not seed:
            seed = random.randint(0, sys.maxsize)
        game = self.game_class(random.Random(seed))

        game.init_board()
        self.players = []
        for bot in self.room.bots:
            # TODO: This is a hack. Remove me sometime.
            if not hasattr(bot, "options"):
                bot.options = {}
            if playback:
                bot.options["debug"] = True

            self.players += [game.create_new_player(bot)]

        screen_cap = []
        if game.TURN_BASED:
            self.current_player = 0
        while game.is_running():
            if playback:
                screen_cap += [game.get_frame()]

            players = self.players
            if game.TURN_BASED:
                players = [self.players[self.current_player]]
                self.current_player = (self.current_player + 1) % len(self.players)
                # TODO: sync screen cap and debug vars.

            for player in players:
                player.run_turn()

            game.do_turn()

        if playback:
            self.room.set_playback(screen_cap)
            for player in self.players:
                self.room.set_bot_debug(player.prog, player.debug_vars)
            return {"seed": int2base(seed, 36)}
        else:  # if score
            return game.get_score()

    def run_for_avg_score(self, times=1, seed=None, func=average):
        """Runs the given game keeping only the scores.

        Args:
            times (int): The number of times to run to get the average score.

        Return:
            The return value the average score for the times runs.
        """
        scores = []
        # TODO: make this able to run in a pool of threads (So it can be run on multiple CPUs)
        for t in range(times):
            scores += [self.__run_for(score=True, seed=seed)]
        return func(scores)

    def run_for_playback(self, seed=None):
        """Runs the given game saving the screen captures.

        Return:
            The return value is a 3-dimensional list the first dimension is time followed by y and x.
        """
        return self.__run_for(playback=True, seed=seed)

    def run(self, seed=None):
        """Will run the game for a user.

        Returns:
        """
        # This import statement is here so pygame is only imported if needed.
        from CYLGame import Display
        game = self.game_class(random.Random(seed))

        charset = Display.CharSet(self.game_class.CHAR_SET, self.game_class.CHAR_WIDTH, self.game_class.CHAR_HEIGHT)
        display = Display.PyGameDisplay(*charset.char_size_to_pix((game.SCREEN_WIDTH, game.SCREEN_HEIGHT)), title=game.GAME_TITLE)

        clock = Display.get_clock()

        frame_buffer = GridFrameBuffer(*charset.pix_size_to_char(display.get_size()), charset=charset)
        frame_updated = True

        game.init_board()
        players = []
        if game.MULTIPLAYER:
            computer_bot_class = game.default_prog_for_computer()
            for _ in range(game.get_number_of_players() - 1):
                players += [game.create_new_player(computer_bot_class())]
        player = game.create_new_player(UserProg())

        while game.is_running():
            clock.tick(FPS)
            for key in display.get_keys():
                # TODO: fix
                player.prog.key = key
                player.run_turn()
                for comp_player in players:
                    comp_player.run_turn()
                game.do_turn()
                frame_updated = True

            if frame_updated:
                game.draw_screen(frame_buffer)
                display.update(frame_buffer)
                frame_updated = False


def run(game_class, avg_game_func=average):
    def serve(args):
        print("I am going to serve")
        from .Server import serve
        serve(game_class, host=args.host, port=args.port, game_data_path=args.dbfile, avg_game_func=avg_game_func)

    def play(args):
        print("Playing...")
        GameRunner(game_class).run(int(args.seed, 36))

    import argparse

    parser = argparse.ArgumentParser(prog=game_class.GAME_TITLE, description='Play ' + game_class.GAME_TITLE + '.')
    subparsers = parser.add_subparsers(help='What do you what to do?')
    parser_play = subparsers.add_parser('play', help='Play ' + game_class.GAME_TITLE + ' with a GUI')
    parser_play.add_argument('-s', '--seed', nargs="?", type=str, help='Manually set the random seed.',
                             default=int2base(random.randint(0, sys.maxsize), 36))
    parser_play.set_defaults(func=play)
    parser_serve = subparsers.add_parser('serve', help='Serve ' + game_class.GAME_TITLE + ' to the web.')
    parser_serve.add_argument('-p', '--port', nargs="?", type=int, help='Port to serve on', default=5000)
    parser_serve.add_argument('-db', '--dbfile', nargs="?", type=str, help='The root path of the game database', default="temp_game")
    parser_serve.add_argument('--host', nargs="?", type=str, help='The mask to host to', default='127.0.0.1')
    parser_serve.set_defaults(func=serve)

    args = parser.parse_args()
    args.func(args)



from setuptools import setup
from distutils.util import convert_path


# Hack to get around version dependency problems
# Taken from: https://stackoverflow.com/a/24517154
main_ns = {}
ver_path = convert_path('CYLGame/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)


setup(name='CYLGame',
      version=main_ns["version"],
      description='UMDCYL Game Framework',
      url='https://github.com/umdlars/CYLGame',
      author='Jonathan Beaulieu',
      author_email='123.jonathan@gmail.com',
      license='MIT',
      packages=['CYLGame'],
      package_dir={'CYLGame': 'CYLGame'},
      package_data={'CYLGame': ['static/*', 'templates/*', 'data/*', 'static/ace/*']},
      install_requires=['future',
                        'flask',
                        'flask-compress',
                        'flask_classful',
                        'Flask-Markdown',
                        'littlepython',
                        'ujson',
                        'msgpack',
                        'gevent'],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      zip_safe=False,
      classifiers=[
            # Complete list of classifiers: http://pypi.python.org/pypi?%3Aaction=list_classifiers
            "Development Status :: 3 - Alpha",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
      ]
      )



from CYLGame.Game import NonGridGame
from CYLGame.Player import Player
from CYLGame.Frame import GameFrame
from CYLGame import GameLanguage
import math
import random

TAU = 2.0 * math.pi;
SEP = "---------------------------------------"

DEBUG = False
def dprint(string):
    if DEBUG:
        print(string)
    

def deg2rad(deg):
    return float(deg) * TAU / 360.0
def rad2deg(rad):
    return float(rad) * 360.0 / TAU

def rotate_point(angle, point):
    cos_ = math.cos(angle)
    sin_ = math.sin(angle)

    newp = [0, 0]
    newp[0] = float(point[0])*cos_ - float(point[1])*sin_
    newp[1] = float(point[0])*sin_ + float(point[1])*cos_

    return newp

def compute_vector(start, end, screen_width, screen_height):
    """
    Computes the vector and distance between two objects, start and end.
    start and end must have a position attribute which is a list/tuple
    of the x and y coord.
    """
    vector = []
    half_width = float(screen_width) / 2.0
    vector.append(end.position[0] - start.position[0])
    if vector[0] > half_width:
        vector[0] -= screen_width
    elif vector[0] < -half_width:
        vector[0] += screen_width

    half_height = float(screen_height) / 2.0
    vector.append(end.position[1] - start.position[1])
    if vector[1] > half_height:
        vector[1] -= screen_height
    elif vector[1] < -half_height:
        vector[1] += screen_height

    return vector[0]**2 + vector[1]**2, vector

class SensorSanitizers(object):
    # sensor sanitize/validate functions
    @staticmethod
    def san_range(r):
        if not r:
            return 0.0

        new_r = float(r)
        if new_r < 0.0:
            new_r = 0.0
        elif new_r > 100.0:
            new_r = 100.0
        return new_r

    @staticmethod
    def san_angle(a):
        if not a:
            return 0.0

        return float(a)

    @staticmethod
    def san_width(w):
        if not w:
            return 0.0

        new_w = float(w)
        if new_w < 0.0:
            new_w = 0.0
        elif new_w > 360.0:
            new_w = 360.0
        return new_w

    @staticmethod
    def san_turret(t):
        if not t:
            return False

        return bool(t)

    import re
    color_re = re.compile(r'#[0-9A-Fa-f]{6}')
    @staticmethod
    def san_color(c):
        if not c:
            return None
        elif SensorSanitizers.color_re.match(c):
            return c
        else:
            return None

class SensorPlayer(Player):

    def add_sensor(self, _range, angle, width, turret):
        """
        Adds a sensor to this player.
        _range: integer [0, 100], the distance that the sensor travels
        angle: integer [0, 360], the direction in degrees the sensor points
        width: integer [0, 360], the width in degrees of the sensor
        turret: bool, if True, the angle is relative to the turret's angle,
        otherwise angle is relative to the tank's angle
        """
        sensor = {}
        sensor["range"] = float(_range)
        sensor["angle"] = deg2rad(angle)
        sensor["width"] = deg2rad(width)
        sensor["turret"] = turret
        sensor["triggered"] = 0
        self.sensors.append(sensor)

    def sensor_calc(self, other, dist_sq, vector, tank_sensor_range):
        """
        See if other is in any of our sensors.
        """
        if self.killer or other.killer:
            return

        # check if they are in our max sensor range
        if dist_sq > (tank_sensor_range + other.radius)**2:
            return
        #print("sensor_calc")
        #print(dist_sq, vector, self.angle)
        # Now calculate sensors
        for i, sensor in enumerate(self.sensors):
            if sensor["range"] <= 0:
                continue

            if sensor["triggered"] & other.obj_type:
                # sensor already firing
                continue

            if dist_sq > (sensor["range"] + other.radius)**2:
                # out of range
                continue

            theta = self.angle + sensor["angle"]
            #print(sensor["angle"])
            if sensor["turret"]:
                theta += self.turret_current

            # do some funky math
            # rotate other's position by theta
            rotated_point = rotate_point(-theta, vector)
            # Sensor is symmetrical, so we only need to consider top
            # quadrants
            rotated_point[1] = abs(rotated_point[1])
            # compute inverse slope of our sensor
            m_s = 1.0 / math.tan(sensor["width"] / 2.0)
            # compute slope to other
            m_r = rotated_point[0] / rotated_point[1] 
            
            # if our inverse slope is less than other, they're inside
            # the arc
            if m_r >= m_s:
                #print("triggered", i)
                sensor["triggered"] |= other.obj_type
                continue

            # Now check if the edge of the arc intersects the tank. Do
            # this just like with firing
            rotated_point = rotate_point(sensor["width"] / -2.0, rotated_point)
            if rotated_point[0] > 0 and abs(rotated_point[1]) < other.radius:
                #print("triggered", i)
                sensor["triggered"] |= other.obj_type



class SensorGame(NonGridGame):
    WEBONLY = True
    NONGRID = True
    OPTIONS = "sensors"

    def do_sensors(self):
        players = self.players
        for player in players:
            for sensor in player.sensors:
                sensor["triggered"] = 0

        for i in range(len(players)):
            if players[i].killer:
                continue

            for j in range(i + 1, len(players)):
                if players[j].killer:
                    continue
                dist_sq, vector = compute_vector(players[i], players[j], self.SCREEN_WIDTH, self.SCREEN_HEIGHT)
                players[i].sensor_calc(players[j], dist_sq, vector, self.MAX_SENSOR_RANGE)
                vector[0] = -vector[0];
                vector[1] = -vector[1];
                players[j].sensor_calc(players[i], dist_sq, vector, self.MAX_SENSOR_RANGE)


